#
# Use this to create valid ID Broker objects through the Provisioning API.
# Must be executed inside the ID Connector container on a school authority server.
#
# devsync -v performance-tests/ 10.200.?.?:/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/
# univention-app shell ucsschool-id-connector
# cd /var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/performance-tests/
# pip install ipython
# ipython
#
# import dev_tools
# await dev_tools.create_provisioning_admin_users("Administrator", "univention", "provisioning.uni.dtr", False)
# await dev_tools.create_x_schools_in_y_school_authorities(4, 20, "provisioning.uni.dtr")
# await dev_tools.create_z_classes_in_x_schools_in_y_school_authorities(5, 4, 20, "provisioning.uni.dtr")
#
import asyncio
import uuid
from typing import Dict, Iterator, List, TypeVar, Union


from ucsschool.kelvin.client import (
    Session as KelvinSession,
    User as KelvinUser,
    UserResource as KelvinUserResource,
)
from ucsschool_id_connector import plugin_loader
from ucsschool_id_connector.models import SchoolAuthorityConfiguration

plugin_loader.load_plugins()

from idbroker.id_broker_client import (  # noqa: E402
    IDBrokerError,
    IDBrokerSchool,
    IDBrokerSchoolClass,
    School,
    SchoolClass,
)

T = TypeVar("T")


def split_list(li: List[T], chunk_size: int) -> Iterator[List[T]]:
    """Split a list into chunks of equal size (except last chunk, which is <= `chunk_size`)."""
    len_li = len(li)
    for i in range(0, len_li, chunk_size):
        yield li[i : i + chunk_size]


async def create_provisioning_admin_users(
    username: str, password: str, host: str, verify: Union[bool, str]
) -> Dict[str, str]:
    """
    Create 20 users (through Kelvin) that can be used to communicate with the provisioning API.

    usernames: `provisioning-TraegerPerf01` to `provisioning-TraegerPerf20`
    passwords: `Univention.Provisioning.01` to `Univention.Provisioning.20`.

    :returns mapping from username to password
    """
    print("Creating 20 provisioning-TraegerPerf?? users...")
    expected_usernames = {f"provisioning-TraegerPerf{num:02d}" for num in range(1, 21)}
    # username=auth.username, password=auth.password, host=auth.host, verify=auth.verify
    async with KelvinSession(username, password, host, verify=verify) as session:
        users = [
            user
            async for user in KelvinUserResource(session=session).search(
                name="provisioning-TraegerPerf*"
            )
        ]
        existing_usernames = {user.name for user in users if user.name in expected_usernames}
        missing_usernames = expected_usernames - existing_usernames
        for username in sorted(missing_usernames):
            print(f"Creating provisioning admin user {username!r}...")
            user = KelvinUser(
                school="DEMOSCHOOL",
                schools=["DEMOSCHOOL"],
                roles=["staff"],
                name=username,
                firstname="provisioning-TraegerPerf",
                lastname=f"number{username[-2:]}",
                password=f"Univention.Provisioning.{username[-2:]}",
                record_uid=username,
                source_uid="provisioning-TraegerPerf",
                session=session,
            )
            await user.save()
            users.append(user)
    print("... done.")
    return {user.name: f"Univention.Provisioning.{user.name[-2:]}" for user in users}


def create_school_authority_configuration(
    school_authority_name: str, password: str, provisioning_host: str
) -> SchoolAuthorityConfiguration:
    return SchoolAuthorityConfiguration(
        name=school_authority_name,
        url=f"https://{provisioning_host}/",
        plugin_configs={
            "id_broker": {
                "username": f"provisioning-{school_authority_name}",
                "password": password,
                "version": 1,
            }
        },
        active=True,
        plugins=["id_broker-users", "id_broker-groups"],
    )


async def create_school_through_provisioning_api(
    school_authority: SchoolAuthorityConfiguration, school_name: str, entry_uuid: str = None
) -> None:
    IDBrokerSchool._share_token = False
    id_broker_school = IDBrokerSchool(school_authority, "id_broker")
    id_broker_school.configuration.verify_ssl = False
    entry_uuid = entry_uuid or str(uuid.uuid4())
    school = School(id=entry_uuid, name=school_name, display_name="Performance Test perm OU")
    print(f"  Creating school {school.name!r}...")
    await id_broker_school.create(school)


async def create_school_class_through_provisioning_api(
    school_authority: SchoolAuthorityConfiguration,
    school_name: str,
    class_name: str,
    entry_uuid: str = None,
) -> None:
    IDBrokerSchoolClass._share_token = False
    id_broker_class = IDBrokerSchoolClass(school_authority, "id_broker")
    id_broker_class.configuration.verify_ssl = False
    entry_uuid = entry_uuid or str(uuid.uuid4())
    school_class = SchoolClass(
        id=entry_uuid,
        name=class_name,
        description="Performance Test perm class",
        school=school_name,
        members=[],
    )
    await id_broker_class.create(school_class)


async def create_x_schools_in_y_school_authorities(
    num_schools: int, num_school_auths: int, provisioning_host: str
) -> None:
    for i in range(1, num_school_auths + 1):
        school_authority = create_school_authority_configuration(
            f"TraegerPerf{i:02}", f"Univention.Provisioning.{i:02}", provisioning_host
        )
        print(f"Creating {num_schools} schools in school authority {school_authority.name!r}...")
        for j in range(1, num_schools + 1):
            name = f"Schule{j:02}"
            print(f"    Creating school class {name!r}...")
            try:
                await create_school_through_provisioning_api(school_authority, name)
                print(f"  School {name!r} created.")
            except IDBrokerError as exc:
                if exc.status == 409:
                    print(f"  School {name!r} exists.")
                    continue
                raise


async def create_z_classes_in_x_schools_in_y_school_authorities(
    num_classes: int, num_schools: int, num_school_auths: int, provisioning_host: str
) -> None:
    """
    Run `create_x_schools_in_y_school_authorities(num_schools, num_school_auths, provisioning_host)`
    before this.
    """

    async def _create(s_a: SchoolAuthorityConfiguration, s_n: str, c_n: str):
        try:
            await create_school_class_through_provisioning_api(s_a, s_n, c_n)
            print(f"    School class {c_n!r} created.")
        except IDBrokerError as _exc:
            if _exc.status == 409:
                print(f"    School class {c_n!r} exists.")
            else:
                raise

    for i in range(1, num_school_auths + 1):
        school_authority = create_school_authority_configuration(
            f"TraegerPerf{i:02}", f"Univention.Provisioning.{i:02}", provisioning_host
        )
        print(f"Creating school classes in school authority {school_authority.name!r}...")
        for j in range(1, num_schools + 1):
            school_name = f"Schule{j:02}"
            print(f"  Creating {num_classes} school classes in school {school_name!r}...")
            class_names = [f"class{k:02}" for k in range(1, num_classes + 1)]
            for chunk in split_list(
                [_create(school_authority, school_name, name) for name in class_names], 4
            ):
                await asyncio.gather(*chunk)
