import datetime
import json
import logging
import time
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union, cast

import jwt
import requests.packages.urllib3
from gevent.lock import BoundedSemaphore
from locust import HttpUser
from locust.clients import HttpSession as LocustClient, ResponseContextManager as LucustCMResponse
from pydantic import BaseModel
from urllib3.exceptions import InsecureRequestWarning

PROVISIONING_SCHOOL_URL = "/ucsschool/apis/provisioning/v1/{school_authority}/schools/{id}"
PROVISIONING_SCHOOL_CLASS_URL = "/ucsschool/apis/provisioning/v1/{school_authority}/classes/{id}"
PROVISIONING_USER_URL = "/ucsschool/apis/provisioning/v1/{school_authority}/users/{id}"
UCSSCHOOL_APIS_TOKEN_URL = "/ucsschool/apis/auth/token"
DATA_DIR1 = Path(__file__).parent.parent / "data"
DATA_DIR2 = Path("/var/lib/id-broker-performance-tests/data")


requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


def get_data_dir() -> Path:
    return DATA_DIR1 if DATA_DIR1.exists() else DATA_DIR2


def get_provisioning_api_admins(path: Path = None) -> Dict[str, str]:
    """
    Get Provisioning API school authority admin users and passwords.

    :param Path path: JSON file from which to read account data or None for default
     :returns mapping username -> password: {username: password}
    """
    path = path or get_data_dir() / "api_admin_users.json"
    res = json.load(path.open("r"))
    logging.info("Loaded %d Provisioning API school authority admin users and passwords.", len(res))
    return res


def get_id_broker_classes(path: Path = None) -> Dict[str, Dict[str, Dict[str, str]]]:
    """
    Get Provisioning API school classes.

    :param Path path: JSON file from which to read class data or None for default
     :returns mappings school authority -> schools -> class -> entryUUID: {sch_auth: {ou: {cl: uuid, ...}}}
    """
    path = path or get_data_dir() / "classes.json"
    res = json.load(path.open("r"))
    s_auths = len(res)
    schools = sum(len(value) for value in res.values())
    classes = sum(sum(len(v) for v in value.values()) for value in res.values())
    logging.info("Loaded %d school classes in %d schools in %d authorities.", classes, schools, s_auths)
    return res


def get_id_broker_schools(path: Path = None) -> Dict[str, Dict[str, str]]:
    """
    Get Provisioning API schools.

    :param Path path: JSON file from which to read schools data or None for default
     :returns mappings school authority -> school -> entryUUID: {sch_auth: {ou: uuid, ...}}
    """
    path = path or get_data_dir() / "schools.json"
    res = json.load(path.open("r"))
    s_auths = len(res)
    schools = sum(len(value) for value in res.values())
    logging.info("Loaded %d schools in %d authorities.", schools, s_auths)
    return res


class ResponseError(Exception):
    ...


class TokenError(Exception):
    ...


class ProvisioningSchool(BaseModel):
    school_authority: str
    id: str
    name: str
    display_name: str = ""

    def __str__(self):
        return f"{self.__class__.__name__}(s_auth={self.school_authority!r}, id={self.id!r}, name={self.name!r})"


class ProvisioningSchoolClass(BaseModel):
    school_authority: str
    id: str
    name: str
    description: str = ""
    school: str
    members: List[str] = []

    def __str__(self):
        return f"{self.__class__.__name__}(s_auth={self.school_authority!r}, id={self.id!r}, name={self.name!r})"


class FastAPIClient(HttpUser):
    abstract = True
    auth_token_url = ""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.auth_username = ""
        self.auth_password = ""
        self._access_token = ""  # do not access this var directly, use the "self.access_token" property
        self._update_time = 0.0
        self._sem_token = BoundedSemaphore()

    def on_start(self) -> None:
        if not self.auth_username or not self.auth_password:
            raise RuntimeError("self.auth_username or self.auth_password not set.")
        _ = self.access_token  # side effect: fetch token

    @property
    def access_token(self) -> str:
        return self._get_access_token(self.client, self.host)

    @access_token.setter
    def access_token(self, token: str) -> None:
        self._set_access_token(token)

    def _head(self, *args, **kwargs) -> LucustCMResponse:
        kwargs.setdefault("headers", {}).update(
            {"accept": "application/json", "Authorization": self.access_token}
        )
        logging.debug("_head() args=%r kwargs=%r", args, kwargs)
        with self.client.head(*args, catch_response=True, verify=False, **kwargs) as response:
            return cast(LucustCMResponse, response)

    def _get_json(
        self, *args, **kwargs
    ) -> Tuple[Union[Dict[str, Any], List[Dict[str, Any]]], LucustCMResponse]:
        kwargs.setdefault("headers", {}).update(
            {"accept": "application/json", "Authorization": self.access_token}
        )
        logging.debug("_get_json() args=%r kwargs=%r", args, kwargs)
        with self.client.get(*args, catch_response=True, verify=False, **kwargs) as response:
            response = cast(LucustCMResponse, response)
            response_json = self._response_to_json(response)
            return response_json, response

    def _get_json_list(self, *args, **kwargs) -> Tuple[List[Dict[str, Any]], LucustCMResponse]:
        res, response = self._get_json(*args, **kwargs)
        if not res:
            res = []
        return res, response

    def _post_json(self, *args, **kwargs) -> Tuple[Dict[str, Any], LucustCMResponse]:
        kwargs.setdefault("headers", {}).update(
            {"accept": "application/json", "Authorization": self.access_token}
        )
        logging.debug("_post_json() args=%r kwargs=%r", args, kwargs)
        with self.client.post(*args, catch_response=True, verify=False, **kwargs) as response:
            response = cast(LucustCMResponse, response)
            response_json = self._response_to_json(response)
            return response_json, response

    def _put_json(self, *args, **kwargs) -> Tuple[Dict[str, Any], LucustCMResponse]:
        kwargs.setdefault("headers", {}).update(
            {"accept": "application/json", "Authorization": self.access_token}
        )
        logging.debug("_put_json() args=%r kwargs=%r", args, kwargs)
        with self.client.put(*args, catch_response=True, verify=False, **kwargs) as response:
            response = cast(LucustCMResponse, response)
            response_json = self._response_to_json(response)
            return response_json, response

    def _delete(self, *args, **kwargs) -> LucustCMResponse:
        kwargs.setdefault("headers", {}).update(
            {"accept": "application/json", "Authorization": self.access_token}
        )
        logging.info("_delete() args=%r kwargs=%r", args, kwargs)
        with self.client.delete(*args, catch_response=True, verify=False, **kwargs) as response:
            response = cast(LucustCMResponse, response)
            if response.status_code != 204:
                msg = f"status code: {response.status_code} Response text: {response.text}"
                response.failure(msg)
                raise ResponseError(msg)
            return response

    def _response_to_json(
        self, response: LucustCMResponse
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        if response.status_code == 401:
            msg = f"status code: {response.status_code} Response text: {response.text}"
            response.failure(msg)
            self.access_token = ""
            raise ResponseError(msg)
        elif response.status_code < 200 or response.status_code > 299:
            msg = f"status code: {response.status_code} Response text: {response.text}"
            response.failure(msg)
            raise ResponseError(msg)
        try:
            return response.json()
        except JSONDecodeError as exc:
            msg = f"Response could not be decoded as JSON: {exc!s}"
            response.failure(msg)
            raise ResponseError(msg) from exc

    @staticmethod
    def _get_token_expiry(token: str) -> datetime.datetime:
        """Get the time at which `token` expires."""
        actual_token = token.rsplit(" ", 1)[-1]
        payload = jwt.decode(actual_token, algorithm="HS256", options={"verify_signature": False})
        exp: str = payload.get("exp")
        ts = int(exp)
        return datetime.datetime.fromtimestamp(ts)

    def _get_access_token(self, http_client, server) -> str:
        self._sem_token.acquire()  # guard access to _access_token
        if not self._access_token or time.time() >= self._update_time:
            logging.info(f"Fetching access token for {self.auth_username!r}...")
            url = f"https://{server}{self.auth_token_url}"
            self._access_token = self._get_fastapi_token(
                http_client, url, self.auth_username, self.auth_password
            )
            self._update_time = time.time() + 3600 - 10  # refresh (shortly before) every hour
        self._sem_token.release()
        return self._access_token

    def _set_access_token(self, token: str) -> None:
        self._sem_token.acquire()  # guard access to _access_token
        self._access_token = token
        self._sem_token.release()

    @staticmethod
    def _get_fastapi_token(client: LocustClient, url: str, username: str, password: str) -> str:
        """Get an access token from the server."""
        headers = {"accept": "application/json", "Content-Type": "application/x-www-form-urlencoded"}
        data = {"username": username, "password": password}
        with client.post(url, data=data, headers=headers, catch_response=True, verify=False) as response:
            response = cast(LucustCMResponse, response)
            try:
                response_json = response.json()
            except JSONDecodeError as exc:
                raise TokenError(
                    f"Fetching token: response could not be decoded as JSON: {exc!s}\nresponse.text: "
                    f"{response.text!r}"
                ) from exc
            try:
                return f"{response_json['token_type']} {response_json['access_token']}"
            except KeyError as exc:
                raise TokenError(
                    f"Fetching token: response did not have expected content: {response_json!r}"
                ) from exc

    def _head_resource(self, url: str, _url_name: str = None) -> LucustCMResponse:
        return self._head(url, name=_url_name)

    def _get_resource(self, url: str, _url_name: str = None) -> Dict[str, Any]:
        res, _ = self._get_json(url, name=_url_name)
        return res

    def _get_resource_list(
        self, url: str, _url_name: str = None, params: Dict[str, Any] = None
    ) -> List[Dict[str, Any]]:
        res, _ = self._get_json_list(url, name=_url_name, params=params)
        return res

    def _create_resource(self, url: str, _url_name: str = None, **body) -> Dict[str, Any]:
        res, _ = self._post_json(url, name=_url_name, json=body)
        return res

    def _delete_resource(self, url: str, _url_name: str = None) -> LucustCMResponse:
        return self._delete(url, name=_url_name)

    def _update_resource(self, url: str, _url_name: str = None, **body) -> Dict[str, Any]:
        res, _ = self._put_json(url, name=_url_name, json=body)
        return res


class ProvisioningClient(FastAPIClient):
    abstract = True
    auth_token_url = UCSSCHOOL_APIS_TOKEN_URL

    def school_exists(self, school_authority: str, obj_id: str) -> bool:
        url = f"https://{self.host}{PROVISIONING_SCHOOL_URL.format(school_authority=school_authority, id=obj_id)}"
        res = self._head_resource(url, _url_name=PROVISIONING_SCHOOL_URL)
        return res.status_code == 200

    def school_get(self, school_authority: str, obj_id: str) -> ProvisioningSchool:
        url = f"https://{self.host}{PROVISIONING_SCHOOL_URL.format(school_authority=school_authority, id=obj_id)}"
        res = self._get_resource(url, _url_name=PROVISIONING_SCHOOL_URL)
        return ProvisioningSchool(school_authority=school_authority, **res)

    def school_create(
        self, school_authority: str, obj_id: str, name: str, display_name: str = ""
    ) -> ProvisioningSchool:
        url_path = PROVISIONING_SCHOOL_URL.format(school_authority=school_authority, id="").rstrip("/")
        url = f"https://{self.host}{url_path}"
        data = {
            "id": obj_id,
            "name": name,
            "display_name": display_name or f"Display name: {name}",
        }
        res = self._create_resource(url, _url_name=PROVISIONING_SCHOOL_URL, **data)
        return ProvisioningSchool(school_authority=school_authority, **res)

    def school_class_get(self, school_authority: str, obj_id: str) -> ProvisioningSchoolClass:
        url_path = PROVISIONING_SCHOOL_CLASS_URL.format(school_authority=school_authority, id=obj_id)
        url = f"https://{self.host}{url_path}"
        res = self._get_resource(url, _url_name=PROVISIONING_SCHOOL_CLASS_URL)
        return ProvisioningSchoolClass(school_authority=school_authority, **res)

    def school_class_exists(self, school_authority: str, obj_id: str) -> bool:
        url_path = PROVISIONING_SCHOOL_CLASS_URL.format(school_authority=school_authority, id=obj_id)
        url = f"https://{self.host}{url_path}"
        res = self._head_resource(url, _url_name=PROVISIONING_SCHOOL_CLASS_URL)
        return res.status_code == 200

    def school_class_create(
        self,
        school_authority: str,
        obj_id: str,
        name: str,
        school: str,
        description: str = "",
        members: List[str] = None,
    ) -> ProvisioningSchoolClass:
        url_path = PROVISIONING_SCHOOL_CLASS_URL.format(school_authority=school_authority, id="").rstrip(
            "/"
        )
        url = f"https://{self.host}{url_path}"
        data = {
            "id": obj_id,
            "name": name,
            "school": school,
            "description": description or f"Description: {name}",
            "members": members or [],
        }
        res = self._create_resource(url, _url_name=PROVISIONING_SCHOOL_URL, **data)
        return ProvisioningSchoolClass(school_authority=school_authority, **res)

    def school_class_update(
        self,
        school_authority: str,
        obj_id: str,
        name: str,
        school: str,
        description: str = "",
        members: List[str] = None,
    ) -> ProvisioningSchoolClass:
        url_path = PROVISIONING_SCHOOL_CLASS_URL.format(school_authority=school_authority, id=obj_id)
        url = f"https://{self.host}{url_path}"
        data = {
            "id": obj_id,
            "name": name,
            "school": school,
            "description": description or f"Description: {name}",
            "members": members or [],
        }
        res = self._update_resource(url, _url_name=PROVISIONING_SCHOOL_URL, **data)
        return ProvisioningSchoolClass(school_authority=school_authority, **res)

    def school_class_delete(self, school_authority: str, obj_id: str) -> None:
        url_path = PROVISIONING_SCHOOL_CLASS_URL.format(school_authority=school_authority, id=obj_id)
        url = f"https://{self.host}{url_path}"
        self._delete_resource(url, _url_name=PROVISIONING_SCHOOL_URL)
